#include <display.h>
#include <avr/interrupt.h>
#include <timer.h>
#include <adc.h>
#include <led.h>
#include <7seg.h>
#include <stdlib.h>

typedef enum {uncovered, number1, number2, number3, number4, number5, number6, number7, number8, bomb, flag, flagAndBomb, covered, cursor } Type; 

static volatile uint8_t press0 = 0;	
static volatile uint8_t press1 = 0;
static uint8_t buttonMode;

static uint8_t fieldArray[18][10];												//Es passen nur 16x8 Zellen auf den Bildschirm, hier aber 18x10 da die Randbetrachtung wegfaellt
static uint8_t mineCount;	
static uint8_t flagsRemaining;													//Wird beim plaztieren/entfernen der Flaggen erniedrigt/erhoeht
static uint8_t gamestate;														//0 = ongoing, 1 = won, 2 = lost

static int8_t cursorX;														
static int8_t cursorY;			

static void init(void);															//Setzt Interrupts und generiert den Zufallszahlenseed anhand der Stellung des Potis

static void drawCell(uint8_t posX, uint8_t posY, Type type);					

static void setMode(void);														//Setzt das Verhalten der Knoepfe anhand des Potis

static void handleButton0(void);												

static void handleButton1(void);

static uint8_t mainMenu(void);													//Wird vor jedem Spielstart aufgerufen, ermittelt Minenzahl und gibt diese zurueck

static uint8_t drawMenuCursor(void);											//Hilfsfunktion fuer das Mainmenu

static void fillFieldArray(void);												//Wird vor jedem Spielstart aufgerufen, fuellt fieldArray zuerst mit bedeckten Zellen und platziert dann die Minen
	
static void floodArea(void);													//Deckt die Zellen in der Umgebung des Cursors auf, bis auf eine Mine gestossen wird
	
static uint8_t checkBombs(uint8_t posX, uint8_t posY);							//Hilfsfunktion fuer floodArea, gibt Anzahl der Bomben in einem 3x3 um den Punkt (posX, posY) zurueck

static void trigger(void);														//Wenn Cursor auf Mine: Gameover,	Bedeckter Zelle: FloodArea,		sonst ignorieren

static void placeFlag(void);													//Platziert Flaggen wenn flagsRemaining > 0, entfernen geht immer

static uint8_t checkWin(void);													//Wenn eine noch nicht geflaggte Mine gefunden wurde, geht das Spiel weiter

static void win(void);

static void gameover(void);


ISR(INT0_vect){
	
	press0 = 1;
}

ISR(INT1_vect){
	
	press1 = 2;
}



static void init(void){

	DDRD &= ~((1 << PORTD2) | (1 << PORTD3));

	PORTD |= ((1 << PORTD2) | (1 << PORTD3));
	
	EICRA &= ~((1 << ISC00) | (1 << ISC10));
	EICRA |= ((1 << ISC01) | (1 << ISC11));
		
	EIMSK |= (1 << INT0) | (1 << INT1);

	sei();
	
	sb_display_enable();
	
	srand(sb_adc_read(POTI) + sb_adc_read(PHOTO));
	srand(rand());
	
}



static void drawCell(uint8_t posX, uint8_t posY, Type type){	
	
	posX = posX * 8;
		
	uint8_t paintArray[8];	
		
	paintArray[0] = 0xff;
	paintArray[1] = 0x81;
	paintArray[6] = 0x81;
	paintArray[7] = 0xff;
	
	switch(type){
		
		case uncovered:		paintArray[1] = 0x81;
							paintArray[2] = 0x81;
							paintArray[3] = 0x81;
							paintArray[4] = 0x81;
							paintArray[5] = 0x81;
							paintArray[6] = 0x81;
							break;
						
		case number1:		paintArray[2] = 0x91;
							paintArray[3] = 0x89;
							paintArray[4] = 0xfd;
							paintArray[5] = 0x81;
							break;		
		
		case number2:		paintArray[2] = 0xe5;
							paintArray[3] = 0xd5;
							paintArray[4] = 0xd5;
							paintArray[5] = 0xc9;
							break;
					
		case number3:		paintArray[2] = 0xc5;
							paintArray[3] = 0xd5;
							paintArray[4] = 0xd5;
							paintArray[5] = 0xa9;
							break;
					
		case number4:		paintArray[2] = 0xbd;
							paintArray[3] = 0xa1;
							paintArray[4] = 0xf9;
							paintArray[5] = 0xa1;
							break;
					
		case number5:		paintArray[2] = 0xdd;
							paintArray[3] = 0xd5;
							paintArray[4] = 0xd5;
							paintArray[5] = 0xf5;
							break;
					
		case number6:		paintArray[2] = 0xfd;
							paintArray[3] = 0xd5;
							paintArray[4] = 0xd5;
							paintArray[5] = 0xf5;
							break;
					
		case number7:		paintArray[2] = 0x85;
							paintArray[3] = 0xe5;
							paintArray[4] = 0x95;
							paintArray[5] = 0x8d;
							break;
		
		case number8:		paintArray[2] = 0xfd;
							paintArray[3] = 0xd5;
							paintArray[4] = 0xd5;
							paintArray[5] =	0xfd;
							break;
			
		case bomb:			paintArray[2] = 0x99;
							paintArray[3] = 0xb5;
							paintArray[4] = 0xad;
							paintArray[5] = 0x99;
							break;
	
		case flag:								
		case flagAndBomb:	paintArray[2] = 0xfd;
							paintArray[3] = 0x9d;
							paintArray[4] = 0x99;
							paintArray[5] = 0x91;
							break;
		
		case covered:		paintArray[1] = 0xff;
							paintArray[2] = 0xc3;
							paintArray[3] = 0xc3;
							paintArray[4] = 0xc3;
							paintArray[5] = 0xc3;
							paintArray[6] = 0xff;
							break;			
			
		case cursor:		paintArray[1] = 0xc3;
							paintArray[2] = 0xa5;
							paintArray[3] = 0x99;
							paintArray[4] = 0x99;
							paintArray[5] = 0xa5;
							paintArray[6] = 0xc3;	
							break;	
	}
	
	for(uint8_t i = 0; i < 8; i++){
			
		sb_display_draw(posY, posX + i, paintArray[i]);
	}
}



static void setMode(void){
	
	sb_led_showLevel(buttonMode = (sb_adc_read(POTI) / 342), 2);
	
	switch( buttonMode ){
		
		case 0: 	sb_7seg_showString("FT");
		break;
		case 1:		sb_7seg_showString("RL");
		break;
		case 2:		sb_7seg_showString("UD");
		break;
	}
}



static void handleButton0(void){
	
	switch ( buttonMode ) {
		
		case 0:	trigger();
		break;
		case 1: if(0 <= cursorX - 1) cursorX--;
		break;
		case 2: if(cursorY + 1 < 8) cursorY++;
		break;
	}
	
	press0 = 0;
}



static void handleButton1(void){
	
	switch ( buttonMode ) {
		
		case 0: placeFlag();
		break;
		case 1: if(cursorX + 1 < 16) cursorX++;
		break;
		case 2: if(0 <= cursorY - 1) cursorY--;
		break;
	}
	
	press1 = 0;
}



static uint8_t mainMenu(void){				
	
	while(1){
		
		press0 = 0;
		press1 = 0;
		
		uint8_t menuPos;
		uint8_t mineCount = 100;
		
		sb_display_fillScreen(0);
		sb_display_showString(0, 15, "Moinsweeper");
		sb_display_showString(2, 25, "Play");
		sb_display_showString(4, 25, "Controls");
		sb_display_showString(6, 25, "Credits");
		
		for(uint8_t i = 0; i < 8; i++){
			for(uint8_t j = 0; j < 2; j++){
				
				drawCell(j*15, i, rand() % 10 + 1);
			}
		}
		
			
		
		do {   menuPos = drawMenuCursor();  } while(press0 == 0);
		
		press0 = 0;
		press1 = 0;
		
		switch (menuPos){
			
			case 0: 		sb_display_fillScreen(0);
							sb_display_showString(0, 15, "Difficulty:");
							sb_display_showString(2, 25, "Easy");
							sb_display_showString(4, 25, "Hard");
							sb_display_showString(6, 25, "Custom");
							
							for(uint8_t i = 0; i < 8; i++){
								for(uint8_t j = 0; j < 2; j++){
											
									drawCell(j*15, i, rand() % 10 + 1);
								}
							}
							
							do { menuPos = drawMenuCursor();  } while(press0 == 0 && press1 == 0);
							
							if(press1) break;
							if(menuPos == 0){
								mineCount = 20;
								break;
							} 
													
							if(menuPos == 1){
								mineCount = 50;
								break;
							}
						
							press0 = 0;
							press1 = 0;
							
							sb_display_fillScreen(0);
							sb_display_showStringWide(3, 20,"Mine Count:");
							
							while(press0 == 0 && press1 == 0){
								
								mineCount = (1023 - sb_adc_read(POTI)) / 10;
								mineCount = ( 99 < mineCount ) ? 99 : mineCount;
								
								sb_7seg_showNumber(mineCount);	
							}
							
							sb_7seg_disable();	
							
							if(press1) mineCount = 100;
								
							break;					
												
							
			case 1:			sb_display_fillScreen(0);
							sb_display_showString(0, 0, "Controls:");
							sb_display_showStringSmall(2, 0, "Potentiometer changes the button");
							sb_display_showStringSmall(3, 0, "mode. HEX left/ right indicates");
							sb_display_showStringSmall(4, 0, "the upper/lower buttons function:");
							sb_display_showStringSmall(5, 0, "> F = set flag,   t = trigger");
							sb_display_showStringSmall(6, 0, "> r = right   ,   l = left");
							sb_display_showStringSmall(7, 0, "> u = up      ,   d = down");

							while(press0 == 0 && press1 == 0){}
								
							press0 = press1 = 0;
								
							sb_display_fillScreen(0);
							sb_display_showString(0, 0, "Controls:");
							sb_display_showStringSmall(2, 0, "Easy   -> 20 Mines");
							sb_display_showStringSmall(3, 0, "Hard   -> 50 Mines");
							sb_display_showStringSmall(4, 0, "Custom -> 0 - 99 Mines");
							sb_display_showStringSmall(5, 0, "You can place as many flags as");
							sb_display_showStringSmall(6, 0, "there are mines. In order to");
							sb_display_showStringSmall(7, 0, "win, all mines must be flagged.");

							while(press0 == 0 && press1 == 0){}
								
							break;
			
			case 2:			sb_display_fillScreen(0);
							sb_display_showString(0, 0, "Credits:");
							sb_display_showStringSmall(2, 0, "Story:        Quentin Tarantino");
							sb_display_showStringSmall(3, 0, "Graphics:     Bob Ross");	
							sb_display_showStringSmall(4, 0, "Sounddesign:  Hans Zimmer");
							sb_display_showStringSmall(5, 0, "Programmer:   Mario Euler");	
							sb_display_showStringSmall(6, 0, "F:            Lukas Witte");		
							sb_display_showStringSmall(7, 0, "random():     Jan Lichtenthaeler ");	
							
							while(press0 == 0 && press1 == 0){}
							break;
		} 
		
		press0 = 0;
		press1 = 0;
		
		if(mineCount < 100) return mineCount;
	}
}




static uint8_t drawMenuCursor(void){
	
		uint8_t menuPos = (1023 - sb_adc_read(POTI)) / 342;
			
			for(uint8_t i = 0; i < 3; i++){
				
				if( i == menuPos){
					
					sb_display_draw(2 + 2 * i, 17, 0x7f);
					sb_display_draw(2 + 2 * i, 18, 0x3e);
					sb_display_draw(2 + 2 * i, 19, 0x1c);
					sb_display_draw(2 + 2 * i, 20, 0x08);
					
				} else {
					
					sb_display_draw(2 + 2 * i, 17, 0);
					sb_display_draw(2 + 2 * i, 18, 0);
					sb_display_draw(2 + 2 * i, 19, 0);
					sb_display_draw(2 + 2 * i, 20, 0);	
				}
			}
			
			sb_timer_delay(400);

			for(uint8_t i = 0; i < 3; i++){
						
					sb_display_draw(2 + 2 * i, 17, 0);
					sb_display_draw(2 + 2 * i, 18, 0);
					sb_display_draw(2 + 2 * i, 19, 0);
					sb_display_draw(2 + 2 * i, 20, 0);
			}

			sb_timer_delay(300);
			
		return menuPos;
}



static void fillFieldArray(void){
	
	for(uint8_t i = 0; i < 18; i++){
		for(uint8_t j = 0; j < 10; j++){
			
			fieldArray[i][j] = covered;
		}
	}
	
	uint8_t i = 0;
	
	while(i < mineCount){

		uint8_t ranX = 1 + rand() % 16;
		uint8_t ranY = 1 + rand() % 8;

		if(fieldArray[ranX][ranY] != bomb){
			
			i++;
			fieldArray[ranX][ranY] = bomb;
		}
	}
}



static void floodArea(void){	
	
	drawCell(cursorX, cursorY, fieldArray[cursorX + 1][cursorY + 1] = checkBombs(cursorX, cursorY));
	
	uint8_t newUncovered = 0;
	
	for(uint8_t r = 1; r < 16; r++){
		
		newUncovered = 0;
		
		for(int8_t i = -r; i < r + 1; i++){					//Geht zuerst in einem Quadrat mit grer werdenden Radius durch, solange bis keine neuen Felder aufgedeckt werden
			for(int8_t j = -r; j < r + 1; j++){
				
				if(fieldArray[cursorX + i + 1][cursorY + j + 1] != covered || cursorX + i < 0 || 15 < cursorX + i || cursorY + j < 0 || 7 < cursorY + j || fieldArray[cursorX + i + 1][cursorY + j + 1] == uncovered) continue;
				
				if(( fieldArray[cursorX + i + 2][cursorY + j + 1] == uncovered || fieldArray[cursorX + i][cursorY + j + 1] == uncovered || fieldArray[cursorX + i + 1][cursorY + j + 2] == uncovered || fieldArray[cursorX + i + 1][cursorY + j] == uncovered)){
					
					drawCell(cursorX + i, cursorY + j, fieldArray[cursorX + i + 1][cursorY + j + 1] = checkBombs(cursorX + i, cursorY + j) );
					newUncovered = 1;
				}
			}
		}
		
		if(!newUncovered) break;
	}

	while(1){
		
		newUncovered = 0;
		
		for(int8_t i = 0; i < 16; i++){						//Geht anschlieend mindestens einmal das ganze Spielfeld durch und solange weiter, bis keine neuen Felder aufgedeckt werden
			for(int8_t j = 0; j < 8; j++){
						
				if(fieldArray[i + 1][j + 1] != covered) continue;
							
				if(( fieldArray[i + 2][j + 1] == uncovered || fieldArray[i][j + 1] == uncovered || fieldArray[i + 1][j + 2] == uncovered || fieldArray[i + 1][j] == uncovered)){
								
					drawCell(i, j, fieldArray[i + 1][j + 1] = checkBombs(i, j) );			
					newUncovered = 1;
				}
			}	
		}
				
		if(!newUncovered) break;
	}
}



static uint8_t checkBombs(uint8_t posX, uint8_t posY){
	
	uint8_t mineCount = 0;
	
	for(int8_t i = -1; i < 2; i++){
		for(int8_t j = -1; j < 2; j++){
			
			if(0 <= posX + i && posX + i < 16 && 0 <= posY + j && posY + j < 8 && (fieldArray[posX + i + 1][posY + j + 1] == bomb || fieldArray[posX + i + 1][posY + j + 1] == flagAndBomb)) mineCount++;
		}	
	}
	return mineCount;
}



static void trigger(void){
	
	switch (fieldArray[cursorX + 1][cursorY + 1]){
		
		case bomb:		gamestate = 2;
						return;
					
		case covered:	floodArea();

		default:		return;
	}
}



static void placeFlag(void){																
	
	switch (fieldArray[cursorX + 1][cursorY + 1]){
		
		case covered:		if(flagsRemaining == 0) return;
							fieldArray[cursorX + 1][cursorY + 1] = flag;
							flagsRemaining--;
							return;
						
		case flag:			fieldArray[cursorX + 1][cursorY + 1] = covered;
							flagsRemaining++;
							return;
					
		case bomb:			if(flagsRemaining == 0) return;	
							fieldArray[cursorX + 1][cursorY + 1] = flagAndBomb;
							flagsRemaining--;
							return;
					
		case flagAndBomb:	fieldArray[cursorX + 1][cursorY + 1] = bomb;
							flagsRemaining++;
							return;
	}
}



static uint8_t checkWin(void){
	
	if(gamestate == 2) return 2;
	
	for(int8_t i = 1; i < 18; i++){
		for(int8_t j = 1; j < 10; j++){
			
			if(fieldArray[i][j] == bomb) return 0;
		}
	}
	
	return 1;
}



static void win(void){
	
	sb_timer_delay(1000);
	
	for(uint8_t i = 0; i < 16; i++){
		for(uint8_t j = 0; j < 8; j++){
				
				drawCell(i, j, number8);
		}
	}
	
	sb_timer_delay(1000);
	

	sb_display_fillScreen(0);

	sb_timer_delay(1000);
	
	sb_display_showStringWide(2, 3,"Congratulations");
	sb_display_showStringWide(5, 35,"You Won!");
	
	press0 = press1 = 0;
		
	while(press0 == 0 && press1 == 0);
	
	
	sb_timer_delay(1000);
	sb_display_fillScreen(0);
	sb_display_showStringWide(2, 22,"Thanks for");
	sb_display_showStringWide(5, 32,"playing!");
	
	press0 = press1 = 0;
		
	while(press0 == 0 && press1 == 0);
	
	sb_timer_delay(1000);
}



static void gameover(void){
	
	for(uint8_t i = 0; i < 16; i++){
		for(uint8_t j = 0; j < 8; j++){
			
			drawCell(i, j, fieldArray[i+1][j+1]);
		}
	}
	
	press0 = press1 = 0;
	
	while(press0 == 0 && press1 == 0);
}



void main(void){
	
	init();
	
	while(1){																						
		
		flagsRemaining = mineCount = mainMenu();
		gamestate = cursorX = cursorY = 0;
		
		fillFieldArray();
		
		for(uint8_t i = 0; i < 16; i++){
			for(uint8_t j = 0; j < 8; j++){
				
				drawCell(i, j, covered);
			}
		}
		
		while(1){															
				

			
			static uint8_t blinker = 0; 
			
			int8_t oldCursorX = cursorX;
			int8_t oldCursorY = cursorY;
			
			if((gamestate = checkWin()) != 0) break;	
			
			setMode();
			
			if(press0) handleButton0();
			if(press1) handleButton1();
	
						
			
			if(oldCursorX != cursorX || oldCursorY != cursorY) 	drawCell(oldCursorX, oldCursorY, (fieldArray[oldCursorX + 1][oldCursorY + 1] != bomb) ? fieldArray[oldCursorX + 1][oldCursorY + 1] : covered);

			if(blinker == 0) {
				blinker = 1;
				drawCell(cursorX, cursorY, cursor);
			} else {
				blinker = 0;
				drawCell(cursorX, cursorY, (fieldArray[cursorX + 1][cursorY + 1] != bomb) ? fieldArray[cursorX + 1][cursorY + 1] : covered);
			}
				

			sb_timer_delay(400);											
		}	
		
		sb_led_setMask(0);
		sb_7seg_disable();
		
		if(gamestate == 1)	{
			
			win();
		} else {
			
			gameover();
		}
		sb_timer_delay(100);
	}
}